
option optimize_for = LITE_RUNTIME;

package TestTags;

message Test1 {

    required int32 i = 1;

}

message Test200 {

    required int32 i = 200;

}

message Test200000 {

    required int32 i = 200000;

}

message TestMax {

    required int32 i = 536870911;

}

