"""
Version information.
"""

# the major version
pyosmium_major = '2.15'
# current release (Pip version)
pyosmium_release = '2.15.0'

# libosmium version shipped with the Pip release
libosmium_version = '2.15.0'
# protozero version shipped with the Pip release
protozero_version = '1.6.4'
# pybind11 version shipped with the Pip release
pybind11_version = '2.2.4'
