
option optimize_for = LITE_RUNTIME;

package TestNested;

message SubSub {
    optional string s = 1;
    optional int32 i = 2;
}

message Sub {
    optional SubSub subsub = 1;
    optional int32 i = 2;
}

message Test {
    optional Sub sub = 1;
    optional int32 i = 2;
}

